
/***********************************************************************************************************************************
*
*	Modul:	avboot.c
*
*	Replacement for Pboot program written by P. Dannegger,	employing the win32 API
*
*	26. Aug. 2006 by jbr	
*
************************************************************************************************************************************/

#include <windows.h>
#include <stdio.h>
#include <unistd.h>
#include <conio.h>


#define	VERSION	"\navboot V1.0, Upload hex-File to ATMega\n"

#define EVER				;;

#define TIMEOUT				100
#define DEFAULT_BAUD		57600

#define ESC					0x1b
#define CR					0x0d

#define XPAGE				512	

#define ONE_BYTE		1

#define AVR_DONE		'a'
#define AVR_CONTINUE	'c'
#define AVR_FAIL			'f'


COMMTIMEOUTS		OldTimeouts;	
HANDLE				ComHandle = NULL;

typedef	int	bool;

char Filename[100];
DWORD	BaudRate= DEFAULT_BAUD;
char Passwd[130] = "\370\360\340\300\374";	// F8,F0,E0,C0,FC
unsigned int Crc=0;

#ifndef EOF
	#define EOF -1
#endif

// Prototypes
int openComPort(char *port);
void setComPortParameter(void);
void giveHelp(void);
int sendChar(unsigned char c);
int sendString(char *str);
int connectATMega(void);
int rxChar(void);
int readValue(  unsigned long *val );
unsigned long readSignature( void );
int readCRC( void );
int readHex( FILE *fp, unsigned int *addr, char *data);
int scanHex( char *str, unsigned int *hexout, int n );
void uploadProgram( char *fname, int EEPROM, int verify );
void buildCRC( unsigned char d );

int main(int argc, char* argv[])
{
	int		key_pressed   = 0;
	char		port[5];

	int c=0;

	if(argc==1)
		giveHelp();
	
	memset(Filename,0,sizeof(Filename));
	memset(port,0,sizeof(port));	

	opterr=0;
	 while (  (c=getopt(argc,argv,"hvf:b:c:") )  != EOF)
	 {
		switch(c)
		{
			case 'f':	strcpy(Filename,optarg);
				break;
			case 'c':	sprintf(port,"COM%d",atoi(optarg));
				break;
			case 'b':	BaudRate=atol(optarg);
				break;
			case 'h':	giveHelp();
				break;
			case 'v':	printf(VERSION);
				break;
			default:
				giveHelp();
				break;
		}
	} 

	// Default COM1
	if(port[0]==0)
		strcpy(port,"COM1");

	if(openComPort(port))
		exit(-1);

	setComPortParameter();
 	printf("Ready to send/receive data. Hit ESC to exit.\n\n");
         
	connectATMega();
	readSignature( );

	uploadProgram( Filename, 0, 0 );

	/***********	not yet working. to be investigated
	if( readCRC() == 0 )
		printf( "CRC pass\n" );
	**********/

	// reset ATMega
	sendString("RES"); sendChar(CR);
	
	while(key_pressed != ESC)
	{
		if (kbhit())
			key_pressed = getch();
	}

	 // Restore time-out parameters
	SetCommTimeouts(ComHandle,&OldTimeouts);
	CloseHandle(ComHandle);
   
	printf("\nDone\n");

	return(0);

} 

int openComPort(char *port)
{   
	
	ComHandle =	CreateFile((LPCTSTR) port,
				GENERIC_READ | GENERIC_WRITE,	
				0,			 						// exclusive access
				NULL,								// no security attributes
				OPEN_EXISTING,             
				FILE_ATTRIBUTE_NORMAL,
				NULL); 
	
	if(ComHandle	 == INVALID_HANDLE_VALUE)
	{
		printf("\nUnable to open < %s >\n\n", port);
		printf("Press any key to exit.");
		_getch();
		return(-1);
	}
	else
		printf("\nPort < %s > opened \n\n", port);	
	
	return 0;
}

void setComPortParameter(void)
{
	DCB dcb;
	COMMTIMEOUTS		myTimeouts;	
	char compar[20];

	GetCommTimeouts(ComHandle,&OldTimeouts);
   
	myTimeouts.ReadIntervalTimeout		= TIMEOUT;
	myTimeouts.ReadTotalTimeoutMultiplier	= TIMEOUT;
	myTimeouts.ReadTotalTimeoutConstant	= TIMEOUT;
	myTimeouts.WriteTotalTimeoutMultiplier	= TIMEOUT;
	myTimeouts.WriteTotalTimeoutConstant	= TIMEOUT;
	
	SetCommTimeouts(ComHandle,&myTimeouts);

	memset(compar,0,sizeof(compar));
	sprintf(compar,"%ld,n,8,1",BaudRate);

	FillMemory(&dcb, sizeof(dcb), 0);
 	dcb.DCBlength = sizeof(dcb);
   	if (!BuildCommDCB(compar, &dcb))
	{
		printf("\nError setting Com-Parameters");
		getch();
	}	
	else
		SetCommState(ComHandle, &dcb);

}

int connectATMega(void)
{
	char waitstr[] = { '|', '/', '-', '\\' };
	int i=1;

	for(EVER)
	{
		printf( "\b%c", waitstr[++i&3] );
		if( kbhit() )
		{
			getch();
			printf("\nAborted\n");
			exit( 254 );
		}
    
		sendString( Passwd ); sendChar(CR);
		if( rxChar(  ) == AVR_DONE )
		{
			printf("\nConnected\n");
			return 0;
  	  	}
 	}
}

int sendString(char *str)
{
	int ret;
	DWORD        bytesWritten = 0;

	ret= WriteFile(ComHandle, str,strlen(str), &bytesWritten, NULL);
	while( *str )
   		 buildCRC( *str++ );
	return ret;
}

int sendChar(unsigned char c)
{
	int ret=0;
	DWORD        bytesWritten = 0;

	ret= WriteFile(ComHandle, &c, 1, &bytesWritten, NULL);
	buildCRC( c );	
	
	return ret;
}

int rxChar(void)
{
	char c;	
	DWORD        bytesRead    = 0;

	if (ReadFile(ComHandle, &c, ONE_BYTE, &bytesRead, NULL) )
	 {
		if (bytesRead == 1)
			return c;
	}
	return 0;
}

unsigned long readSignature( void )
{

	unsigned long signature = 0;
	char *dev;

	sendString( "RSI" ); sendChar(CR);

	while( readValue( &signature ) != 1 )
		 ;
		
	if( signature )
	{

		switch( signature )
		{
			case 0x1E9307: dev = "ATMega8";	break;
			case 0x1E9403: dev = "ATMega16";	break;
			case 0x1E9404: dev = "ATMega162";	break;
			case 0x1E9501: dev = "ATMega323";	break;
			default: dev = "";
		}
		printf( "Signature: %lX Device: %s\n", signature, dev );
	}
	return signature;
}

int readValue(  unsigned long *val )
{

	unsigned long newval = 0;
	unsigned int i, v = 0;

	for(EVER)
	{
		i = rxChar( );
		switch( i )
		{
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9': i += 'A' - '0' - 10;
			case 'A':
			case 'B':
			case 'C':
			case 'D':
			case 'E':
			case 'F': newval = newval * 16 + i - 'A' + 10;
				v = 1;
				break;

			case AVR_CONTINUE:
				return 2;

			case AVR_DONE:
				*val = newval;
				return v;		// 0, 1

			case AVR_FAIL:
				*val = newval;
				return -2 - v;	// -2, -3

			default:	return -1;
		}

	}
}

void buildCRC( unsigned char d )
{

	int i;

	Crc ^= d;
	for( i = 8; i ; i-- )
		Crc = (Crc >> 1) ^ ((Crc & 1) ? 0xA001 : 0 );
  
}

int readCRC( void )
{

	unsigned long val;
	unsigned int c;
	int i=0;

	sendString( "CRC" ); sendChar(CR);

	c = Crc;
	Crc = 0;

	for(EVER)
	{  
		i=	readValue( &val ) ;
		printf("\n i = %d, c=%d   val=%ld \n",i, c, val);

 		switch (i)
		{
			case 1:
			if( val == c )		// CRC read
				return 0;
			return -2;
 			
			case -1:
				return -1;		// Timeout
		}
	}
}

int scanHex( char *str, unsigned int *hexout, int n )
{
 	unsigned int hex = 0, x = 0;

	for(; n; n--)
	{
		x = *str;
		if( x >= 'a' )
			x += 10 - 'a';
		else if( x >= 'A' )
			x += 10 - 'A';
		else
			x -= '0';
 		
		if( x >= 16 )
			break;
		
		hex = hex * 16 + x;
		str++;
  	}
 	*hexout = hex;
	return n;					// 0 if all digits read
}


int readHex( FILE *fp, unsigned int *addr,  char *data)
{  
	
	/* Return value:	1..255	number of bytes
					0	end or segment record
					 -1	file end
					 -2	error or no HEX-File
	 */

	char hexline[524];				// intel hex: max 255 byte	
	char * hp =  hexline;
	unsigned int byte, num;
	int i;


 	if( fgets( hexline, 524, fp ) == NULL)
		return -1;		// end of file

	if( *hp++ != ':' )
		return -2;		// no hex record

	if( scanHex( hp, &num, 2 ))
		return -2;		// no hex number

	hp += 2;
	if( scanHex( hp, addr, 4 ))
		return -2;
  	hp += 4;
	
	if( scanHex( hp, &byte, 2 ))
		return -2;

	if( byte != 0 )					
		return 0;		// end or segment record
	
	for( i = num; i--; )
	{
		hp += 2;
		if( scanHex( hp, &byte, 2 ))
			return -2;
	
		*data++ = byte;
	}
	return num;
}

void uploadProgram( char *fname, int EEPROM, int verify )
{
  
	FILE *fp;
	
	unsigned int firstaddr = 65536;
	unsigned int lastaddr = 0;
	unsigned int addr;
	int i;
	char  *data;
	char s[255];
	unsigned char d1;
	unsigned long val;
	char * pv = "Program";

	data = malloc( 65536L );
	if( data == NULL )
	{
		puts("Memory allocation error !" );
		return;
	}
	memset( data, 0xFF, 65536L );

	if( NULL == ( fp = fopen( fname, "rb" ) ) )
	{
		printf("File %s open failed !\n", fname);
		return;
	}

	while( (i = readHex( fp, &addr, s )) >= 0 )
	{
		if( i )
		{
			memcpy( data + addr, s, i );
			if( firstaddr > addr )
				firstaddr = addr;
			addr += i;

			if( lastaddr < addr-1 )
				lastaddr = addr-1;
			
			addr++;
		}
	}
	fclose( fp );

	if( verify )
		pv = "Verify";

	printf( "%s %s: %04X - %04X", pv, fname, firstaddr, firstaddr);

	if( EEPROM )
		sendString( "DEM8" );
	else
		sendString( "DM8" );
	sendChar(CR);

	if( readValue( &val ) != 0 )
		printf( " Programmer Error 1\n" );
	else
	{
		if( verify )
			sprintf( s, "VER%04X", firstaddr );
		else
			sprintf( s, "PRO%04X", firstaddr );
		
		sendString( s ); sendChar(CR);

		for( addr = firstaddr;  ; addr++ )
		{
			d1 = data[addr];
			sendChar( d1 );
			if( d1 == 0xA5 )
				sendChar( 0x00 );				// A5,00 = A5
     
	  		if( (addr & 0x7F) == 0x7F )
				printf( "\b\b\b\b%04X", addr + 1 );
      		
			if( (addr & XPAGE-1) == XPAGE - 1 )
			{
				i = readValue(&val );
				if( i != 2 )				// 2 = continue
				{
					printf( " Error at address: %lX\n", val );
					break;
				}
			}

			if( addr == lastaddr )
			{
				sendChar( 0xA5);
				sendChar( 0xFF );				// A5,FF = End
				printf( "\b\b\b\b%04X\n", addr );
				
				i = readValue(&val );

				if( i == 0 )
					printf( "%s successful", pv );
				else
					printf( " Error at address: %lX\n", val );
	
				break;
			}
		}
	}
	free( data );
	printf("\n");
}

void giveHelp(void)
{
	printf(VERSION);
	printf("use:   avboot  -c  -b  -f\n");
	printf("\n-f Filename");
	printf("\n-c 1-4 Comport");
	printf("\n-b Baudrate\n");
	printf("\ndefaults: Baud=57600, COM1");
	printf("\nExample:  avboot -c2 -b19200 -fTestfile.hex\n\n"); 
	
	exit(0);
}
